/*  dem.c -- functions for reading USGS Digital Elevation Model files  */

#include "config.h"
#include "dem.h"

static Dem *tryReadDEM(FILE *fp);
static Em *dem2em(Dem *dem);
static int getDEMInt(FILE *fp);
static int getDEMInt6(FILE *fp);
static double getDEMDouble12(FILE *fp);
static double getDEMDouble24(FILE *fp);

extern int nogzip;

Em *readEM(char *filename)
{
  Dem *dem;
  Em *em;

  dem = readDEM(filename);
  em = dem2em(dem);
  free(dem);
  return(em);
}


Dem *readDEM(char *filename)
{
  Dem *dem;
  FILE *f;

#if defined(GZCAT) && defined(POPEN) && defined(PCLOSE)

  /* Handle gzipped files as well as regular files */
  char cmd[256];
  FILE *p;

  if (nogzip == 0) {		/* skip if nogzip flag is set */
    dem = NULL;
    if (filename == NULL) {
      sprintf(cmd,"%s",GZCAT);
    } else {
      sprintf(cmd,"%s %s",GZCAT,filename);
    }
    if ((p=POPEN(cmd,"r")) != NULL) {
      dem = tryReadDEM(p);
      PCLOSE(p);
    }
    if (dem != NULL) {
      return dem;
    } else {
      fprintf(stderr,"Bad file or error executing \"%s\".\n",cmd);
      if (filename == NULL) return NULL;	/* try at stdin failed */
    }
  }
#endif

  /* Cannot unzip file so try to read it as a DEM. */
  if (filename == NULL) {
    dem = tryReadDEM(stdin);
    return dem;
  }
  f = fopen(filename,"rb");
  if (f == NULL) {
    fprintf(stderr, "Can't open %s for read.\n", filename);
    return NULL;
  }
  dem = tryReadDEM(f);
  fclose(f);
  return dem;
}


static Em *dem2em(Dem *dem)
{
  Em *em;
  int i, j, length;
  unsigned short *p;
  UV min;
	
  if (dem == NULL) return NULL;

  em = (Em *)malloc(sizeof(Em));
  if (!em) nospace();
	
  em->width = dem->cols;
  em->height = dem->rows;
  length = em->width * em->height;
  em->data = (unsigned short *)malloc(sizeof(short) * length);
  if (!em->data) nospace();
	
  em->dz2m = dem->dz2m;
  em->dzoffset = dem->dzoffset;
  em->dzmax = dem->dzmax;

  for (i = 0; i < 4; i++) {
    if (i == 0) {
      min.u = dem->corner[i].u;
      min.v = dem->corner[i].v;
    } else {
      min.u = MIN(min.u, dem->corner[i].u);
      min.v = MIN(min.v, dem->corner[i].v);
    }
  }
  em->swXY.u = floor((dem->corner[0].u - min.u) / dem->resolution[0]);
  if (em->swXY.u < 0.0)
    em->swXY.u = 0.0;
  em->swXY.v = floor((dem->corner[0].v - min.v) / dem->resolution[1]);
  if (em->swXY.v < 0.0)
    em->swXY.v = 0.0;
  em->nwXY.u = floor((dem->corner[1].u - min.u) / dem->resolution[0]);
  if (em->nwXY.u < 0.0)
    em->nwXY.u = 0.0;
  em->nwXY.v = floor((dem->corner[1].v - min.v) / dem->resolution[1]);
  if (em->nwXY.v > dem->rows - 1)
    em->nwXY.v = dem->rows - 1;
  em->neXY.u = floor((dem->corner[2].u - min.u) / dem->resolution[0]);
  if (em->neXY.u > dem->cols - 1)
    em->neXY.u = dem->cols - 1;
  em->neXY.v = floor((dem->corner[2].v - min.v) / dem->resolution[1]);
  if (em->neXY.v > dem->rows - 1)
    em->neXY.v = dem->rows - 1;
  em->seXY.u = floor((dem->corner[3].u - min.u) / dem->resolution[0]);
  if (em->seXY.u > dem->cols - 1)
    em->seXY.u = dem->cols - 1;
  em->seXY.v = floor((dem->corner[3].v - min.v) / dem->resolution[1]);
  if (em->seXY.v < 0.0)
    em->seXY.v = 0.0;
	
  if (dem->reference_system == 1 && dem->ground_unit == 2) {
		
    /* coordinate system in UTM with meter ground units */
		
    utm2ll(dem->zone, dem->corner[0], &(em->swUV));
    utm2ll(dem->zone, dem->corner[1], &(em->nwUV));
    utm2ll(dem->zone, dem->corner[2], &(em->neUV));
    utm2ll(dem->zone, dem->corner[3], &(em->seUV));
		
  } else {
		
    /* Hope the coordinate system is in arcseconds */
		
    em->swUV.u = dem->corner[0].u;
    em->swUV.v = dem->corner[0].v;
    em->nwUV.u = dem->corner[1].u;
    em->nwUV.v = dem->corner[1].v;
    em->neUV.u = dem->corner[2].u;
    em->neUV.v = dem->corner[2].v;
    em->seUV.u = dem->corner[3].u;
    em->seUV.v = dem->corner[3].v;
  }
	
  p = em->data;
  for(j = 0; j < dem->rows; ++j) {
    for(i = 0; i < dem->cols; ++i) {
      *p++ = dem->data[i][j];
    }
  }
  return(em);
}




static Dem *tryReadDEM(FILE *fp)
{
  Dem *dem;
  int i,j,k;
  int col;
  int row_id, col_id;
  int colsInProfiles, numProfiles;
  UV min, max;
  int firstIndex;
  int dzoffset;
  int truncated;
  int tmp;
  unsigned short dzmax;

  dem = (Dem *)malloc(sizeof(Dem));
  if (dem == NULL) {
    fprintf(stderr,"Can't allocate space for Dem structure.\n");
    return(NULL);
  }
	
  /* A records */
	
  /* A 1 */
  if (!fgets(dem->name, 145, fp)) {
    fprintf(stderr, "Empty input file.\n");
    free(dem);
    return NULL;
  }
	
  /* A 2 */
  dem->level = getDEMInt6(fp);
	
  /* A 3 */
  dem->elevation_pattern = getDEMInt6(fp);
	
  /* A 4 */
  dem->reference_system = getDEMInt6(fp);
	
  /* A 5 */
  dem->zone = getDEMInt6(fp);
	
  /* A 6 */
  for (i = 0; i < 15; i++) {
    dem->map_projection_parameter[i] = getDEMDouble24(fp);
  }
	
  /* A 7 */
  dem->ground_unit = getDEMInt6(fp);
	
  /* A 8 */
  dem->elevation_unit = getDEMInt6(fp);
  if (dem->elevation_unit == 1) {
    dem->dz2m = METERS_PER_FEET;
  } else {
    dem->dz2m = 1.0;
  }
	
  /* A 9 */
  dem->n_polygon = getDEMInt6(fp);
	
  /* A 10 */
  for (i = 0; i < 4; i++) {
    dem->corner[i].u = getDEMDouble24(fp);
    dem->corner[i].v = getDEMDouble24(fp);
    if (i == 0) {
      min.u = dem->corner[i].u;
      max.u = dem->corner[i].u;
      min.v = dem->corner[i].v;
      max.v = dem->corner[i].v;
    } else {
      min.u = MIN(min.u, dem->corner[i].u);
      max.u = MAX(max.u, dem->corner[i].u);
      min.v = MIN(min.v, dem->corner[i].v);
      max.v = MAX(max.v, dem->corner[i].v);
    }
  }
	
  /* A 11 */
  dem->minElevation = getDEMDouble24(fp);
  dem->maxElevation = getDEMDouble24(fp);

  /* A 12 */
  dem->axis_angle = getDEMDouble24(fp);		/* assumed to be 0.0 */
	
  /* A 13 */
  dem->accuracy = getDEMInt6(fp);
	
  /* A 14 */
  for (i = 0; i < 3; i++) {
    dem->resolution[i] = getDEMDouble12(fp);
  }
  if (dem->resolution[2] == 0) {
    fprintf(stderr,"Error reading DEM file.\n");
    free(dem);
    return NULL;
  }
  dem->dz2m *= dem->resolution[2];	/* scale by accuracy */
  dzoffset = (int)(dem->minElevation/dem->resolution[2])-1;
  if (dzoffset < -32767) dzoffset=0-1;
	
  /* A 15 */
  colsInProfiles = getDEMInt6(fp);
  if (colsInProfiles != 1) {
    fprintf(stderr, "Cannot handle more than one column profiles.\n");
    free(dem);
    return NULL;
  }

  numProfiles = getDEMInt6(fp);
  if (numProfiles > MAXCOLS) {
    fprintf(stderr, "Too many profiles (%d) in data set.\n",numProfiles);
    free(dem);
    return NULL;
  }
	
  /* A remaining (ignored) */
  for (i = 864; i < 1024; i++)
    if (getc(fp) == '\n')
      break;
		
  /* B records */
		
  dem->cols = numProfiles;
  dem->rows = (int)(1 + (max.v - min.v)/dem->resolution[1]);
		
  if (dem->rows > MAXROWS) {
    fprintf(stderr, "Too many rows (%d) in DEM.\n",dem->rows);
    free(dem);
    return NULL;
  }

  truncated = 0;
  dzmax = 0;
  for (col = 0; col < numProfiles; ++col) {
    for (i = 0; i < dem->rows; ++i) {
      dem->data[col][i] = 0;
    }
			
    /* B 1 */
    row_id = getDEMInt(fp);
    col_id = getDEMInt(fp);
    if (col_id != col + 1 || row_id != 1) {
      fprintf(stderr, "Profile %d claims to be %d.\n", col + 1, col_id);
      free(dem);
      return NULL;
    }
			
    /* B 2 */
    dem->rowsInProfile[col] = getDEMInt6(fp);
    dem->colsInProfile[col] = getDEMInt6(fp);
			
    /* B 3 */
    dem->firstUVInProfile[col].u = getDEMDouble24(fp);
    dem->firstUVInProfile[col].v = getDEMDouble24(fp);
    firstIndex = (int)((dem->firstUVInProfile[col].v - min.v)/dem->resolution[1]);
			
    /* B 4 */
    dem->firstElevationInProfile[col] = getDEMDouble24(fp);
			
    /* B 5 */
    dem->minElevationInProfile[col] = getDEMDouble24(fp);
    dem->maxElevationInProfile[col] = getDEMDouble24(fp);
			
    /* B 6 */
#ifdef DEBUG
    fprintf(stderr, "Profile %d starts with row (%lf - %lf)/%lf = %d.\n",
	    col + 1, dem->firstUVInProfile[col].v, min.v, dem->resolution[1],
	    firstIndex);
#endif
			
    for (i = firstIndex; i < firstIndex + dem->rowsInProfile[col]; i++) {
      if (i >= dem->rows) {
	fprintf(stderr, "Too many rows in DEM profile %d.\n", col + 1);
	free(dem);
	return NULL;
      }

      /* WARNING some DEM files have record delimiters (like '\n')
	 that may break a single elevation into two.  We do not handle
	 these types of DEM files.
	 */
      tmp = getDEMInt(fp);

      if (tmp != -32767) {		/* -32767 ==> no data */
	if (tmp < dzoffset+1) {		/* the min elevation was wrong */
#ifndef NDEBUG
	  fprintf(stderr,"Rescaling for new minimum elevation.\n");
#endif
	  for (j = 0; j <= col; ++j) {
	    for (k = 0; k < dem->rows; ++k) {
	      if (dem->data[j][k] != 0) {
		if (dem->data[j][k] + dzoffset-tmp+1 > USHRT_MAX) {
		  truncated = 1;
		  dem->data[j][k] = USHRT_MAX;
		} else {
		  dem->data[j][k] += (unsigned short)(dzoffset-tmp+1);
		}
		dzmax = MAX(dzmax,dem->data[j][k]);
	      }
	    }
	  }
	  dzoffset = tmp-1;
	}

	if (tmp-dzoffset > USHRT_MAX) {
	  truncated = 1;
	  dem->data[col][i] = USHRT_MAX;
	} else {
	  dem->data[col][i] = (unsigned short)(tmp-dzoffset);
	}
	dzmax = MAX(dzmax,dem->data[col][i]);
      }
    }
			
  }
		
  if (truncated) {
    fprintf(stderr,"**Warning** Some elevations have been truncated.\n");
  }
  dem->dzoffset = dzoffset;
  dem->dzmax = dzmax;

  /* C record */
		
  if (dem->accuracy != 0) {
			
    /* C 1 */
    dem->C2 = getDEMInt6(fp);
			
    /* C 2 */
    dem->C2x = getDEMInt6(fp);
    dem->C2y = getDEMInt6(fp);
    dem->C2z = getDEMInt6(fp);
			
    /* C 3 */
    dem->C2sample_size = getDEMInt6(fp);
			
    /* C 4 */
    dem->C5 = getDEMInt6(fp);
			
    /* C 5 */
    dem->C5x = getDEMInt6(fp);
    dem->C5y = getDEMInt6(fp);
    dem->C5z = getDEMInt6(fp);
			
    /* C 6 */
    dem->C5sample_size = getDEMInt6(fp);
  }
		
  return(dem);
}


static void getDEMD2E(int n, char *buf, FILE *fp)
{
  int i;
	
  for (i = 0; i < n; ++i) {
    buf[i] = fgetc(fp);
    if (buf[i] == 'D') {
      buf[i] = 'E';
    }
  }
  buf[n] = '\0';
}



/* ? characters */
static int getDEMInt(FILE *fp)
{
  int i;
	
  fscanf(fp, "%d", &i);

  return i;
}


static void getDEMNoCR(int n, char *buf, FILE *fp)
{
  int i;
	
  for (i = 0; i < n; ++i) {
    buf[i] = fgetc(fp);
    if (buf[i] == '\n') {
      --i;
    }
  }
  buf[n] = '\0';
	
}


/* 6 characters */
static int getDEMInt6(FILE *fp)
{
  char buf[7];
	
  getDEMNoCR(6, buf, fp);
  return atoi(buf);
}


/* 12 characters */
static double getDEMDouble12(FILE *fp)
{
  char buf[13];
	
  getDEMD2E(12, buf, fp);
  return atof(buf);
}



/* 24 characters */
static double getDEMDouble24(FILE *fp)
{
  char buf[25];
	
  getDEMD2E(24, buf, fp);
  return atof(buf);
}
