/*  emit.c -- draw a surface  */

#ifdef WIN32
#include <windows.h>
#endif
#include "topo.h"
#include <GL/gl.h>

static void emitDFS(Hier *hp, TriIndex *t, Coord2d p1, Coord2d p2, Coord2d p3);
static void emitColor(int x, int y, int z);
static void getNormal(int,int,double,int,int,double,int,int,double,double*,double*,double*);
static void emitTri(Hier *hp, TriIndex *t, Coord2d p1, Coord2d p2, Coord2d p3);



static int surfType;		/* surface type: WIRE2D, WIRE3D, or FILL3D */
static double zscale;


/*  emitSurface(hp, type) -- draw surface from triangle hierarchy  */

void emitSurface(Hier *hp, int type)
{
  surfType = type;
  zscale = vexagger * m2grid * gem->dz2m;
  numDisplayTri = 0;
  emitDFS(hp, &rootL, sw, ne, nw);	/* NW corner of square */
  emitDFS(hp, &rootR, ne, sw, se);	/* SE corner of square */
}



/*  emitDFS(hp, t, p1, p2, p3) -- emit triangle, possibly subdivided  */

static void emitDFS(Hier *hp, TriIndex *t, Coord2d p1, Coord2d p2, Coord2d p3)
{
  Coord2d mid;
  TriIndex lt,rt;

  if (t->s) {
    if (ONSURFACE(hp,t)) {
      emitTri(hp, t, p1, p2, p3);
    } else {
		
      /*   assert(!ISLEAF(hp, t)); */
		
      midpoint(p1, p2, mid);
      leftChild(t,&lt);
      rightChild(t,&rt);
      emitDFS(hp, &lt, p3, p1, mid);
      emitDFS(hp, &rt, p2, p3, mid);
    }
  }
}



/*  emitTri(hp, t, p1, p2, p3) -- emit one triangle  */

static void emitTri(Hier *hp, TriIndex *t, Coord2d p1, Coord2d p2, Coord2d p3)
{
  int x1,x2,x3,y1,y2,y3,z1,z2,z3;
  double sz1, sz2, sz3;
  double nx, ny, nz;

  assert(t->s);

  if (INROD(hp,t) && (!cullFOV || INFOV(hp,t))) {
			
    x1=p1[0];
    y1=p1[1];
    z1=EMDATA(gem, x1, y1);
    sz1 = zscale*z1;

    x2=p2[0];
    y2=p2[1];
    z2=EMDATA(gem, x2, y2);
    sz2 = zscale*z2;

    x3=p3[0];
    y3=p3[1];
    z3=EMDATA(gem, x3, y3);
    sz3 = zscale*z3;

    getNormal(x1,y1,sz1,x2,y2,sz2,x3,y3,sz3,&nx,&ny,&nz);

    /* backface culling */
    if ( !cullBackface || 
	 (x1-eyeEmx)*nx + (y1-eyeEmy)*ny + (sz1-eyeEmz)*nz <= 0.0 ) {

      numDisplayTri++;

      if (surfType == WIRE2D) {

	glBegin(GL_LINE_LOOP);
	glVertex2i(x1, y1);
	glVertex2i(x2, y2);
	glVertex2i(x3, y3);
	glEnd();
			
      } else if (surfType == FILL2D) {
			
	glBegin(GL_POLYGON);
			
	switch(INFOV(hp,t)) {
	case ALLINFOV:
	  glColor3d(1.0, 1.0, 0.0);
	  break;
	case PARTINFOV:
	  glColor3d(1.0, 0.0, 0.0);
	  break;
	default:
	  glColor3d(0.5, 0.5, 0.5);
	}
			
	glVertex2i(x1, y1);
	glVertex2i(x2, y2);
	glVertex2i(x3, y3);
	glEnd();
			
	glBegin(GL_LINE_LOOP);
	glColor3d(0.0, 0.0, 0.0);
	glVertex2i(x1, y1);
	glVertex2i(x2, y2);
	glVertex2i(x3, y3);
	glEnd();
			
      } else {	/* WIRE3D or FILL3D */

	if (surfType == WIRE3D) {
	  glBegin(GL_LINE_LOOP);
	  glVertex3d((double)x1, (double)y1, sz1);
	  glVertex3d((double)x2, (double)y2, sz2);
	  glVertex3d((double)x3, (double)y3, sz3);
	  glEnd();
	} else {
	  glBegin(GL_POLYGON);
	  glNormal3d(nx, ny, nz);
	  emitColor(x1, y1, z1); glVertex3d((double)x1, (double)y1, sz1);
	  emitColor(x2, y2, z2); glVertex3d((double)x2, (double)y2, sz2);
	  emitColor(x3, y3, z3); glVertex3d((double)x3, (double)y3, sz3);
	  glEnd();
	}
      }
    }
  }
}


static void getNormal(int x1, int y1, double sz1,
		      int x2, int y2, double sz2,
		      int x3, int y3, double sz3,
		      double *nxp, double *nyp, double *nzp)
{
  double px, py, pz, len;
	
  double v1x = x2 - x1,  v1y = y2 - y1,  v1z = sz2 - sz1;
  double v2x = x3 - x2,  v2y = y3 - y2,  v2z = sz3 - sz2;
	
  px = v1y * v2z - v1z * v2y;
  py = v1z * v2x - v1x * v2z;
  pz = v1x * v2y - v1y * v2x;
	
    /* Normalize */
  len = px * px + py * py + pz * pz;
  len = sqrt(len);
  *nxp = px / len;
  *nyp = py / len;
  *nzp = pz / len;
	
}



/*  emitColor(x, y, z) -- emit color depending on location and altitude  */

static void emitColor(int x, int y, int z)
{
  if (colorLayer == T_CONTOURS){	/* base the color on the elevation */
    Color *c;

    if (z == 0) {
      glColor3d(0.4, 0.0, 0.0);		/* use dark red for non-data */
    } else if (z <= high_water && z >= low_water) {
      glColor3d(0.0, 0.0, 0.4);		/* use dark blue for sea level */
    } else {
      c = elevMap + (int)(z * dz2contour);
      glColor3d(c->red, c->green, c->blue);
    }
			
  } else if (colorLayer == T_GRAY) {	/* gray with slight tan tint */
			
    glColor3d(0.6, 0.6, 0.55);
			
  } else {				/* use color from PPM files */

    glColor3d(RED(clayer[colorLayer], x, y) / 255.0,
	      GREEN(clayer[colorLayer], x, y) / 255.0,
	      BLUE(clayer[colorLayer], x, y) / 255.0);
  }			
}
