/*  view.c -- display perspective view  */

#ifdef WIN32
#include <windows.h>
#endif
#include "topo.h"
#include <GL/gl.h>
#include <GL/glut.h>

static void refreshPsp(void);

static int newSurface = 1;	/* does surface need to be recomputed? */

static double cosHalfAngle2 = 0.85355339;
static double angle = 45.0;

#define NEAR_EYE_UP 1
#define NEAR_EYE_DOWN 0.1
static double myfar;



/*  viewChanged() -- take note that view has changed, so surface will, too  */
void viewChanged(void)
{

#ifdef TESTING
  if (doingLoop) {
    newSurface = 1;
    if (doRedraw) {
      glutSetWindow(ovhWin);
      refreshOvh();
      glutSetWindow(pspWin);
      refreshPsp();
      glutSetWindow(ovhWin);
    } else {
      buildSurface();
    }
    return;
  }
#endif

  newSurface = 1;
  glutSetWindow(ovhWin);
  glutPostRedisplay();
  glutSetWindow(pspWin);
  glutPostRedisplay();
}



/*  buildSurface() -- rebuild the surface if newSurface is set  */

void buildSurface(void)
{
  int x, y, z;

  if (!newSurface)		/* if nothing's changed, don't rebuild */
    return;

  /* compute eye height depending on location and eyeUp flag */

  if (eyeUp) {

    /* eye is up: EYE_HEIGHT meters above highest peak */
    eyeEmz = vexagger * m2grid * (gem->dzmax*gem->dz2m + EYE_HEIGHT);

  } else {

    /* eye down: just above max height of gem cell corners */
    x = eyeEmx;
    y = eyeEmy;
    z = zval(x, y);
    z = MAX(z, zval(x + 1, y));
    z = MAX(z, zval(x + 1, y + 1));
    z = MAX(z, zval(x, y + 1));
    eyeEmz = vexagger * m2grid * (z*gem->dz2m + EYE_HEIGHT);
  }

  /* compute reference height and vector */

  refEmz=vexagger*m2grid*(zval((int)refEmx,(int)refEmy)*gem->dz2m+REF_HEIGHT);

  eye2refx = refEmx - eyeEmx;
  eye2refy = refEmy - eyeEmy;
  eye2refz = refEmz - eyeEmz;
  inAngleThresh = (eye2refx * eye2refx + eye2refy * eye2refy) * cosHalfAngle2;

  /* build the surface */

  errSurface(hierarchy);

  newSurface = 0;
}



/*  zval(x, y) -- return z value from gem, or 0 if out of bounds  */

int zval(int x, int y)
{
  if (x < 0 || y < 0 || x >= gem->width || y >= gem->height)
    return 0;
  else
    return EMDATA(gem, x, y);
}



/*  openPspWin() -- open perspective view window  */

void openPspWin(void)
{
  GLint tmp;
  GLfloat tmpf;

  GLfloat light_specular[] = { 0.2F, 0.2F, 0.2F, 1.0F };
  GLfloat light_diffuse[] = { 1.0F, 1.0F, 1.0F, 1.0F };
  GLfloat light_ambient[] = { 0.0F, 0.0F, 0.0F, 1.0F };
  GLfloat light_position[] = {0.0F, 1.0F, 2.0F, 0.0F};/*high in northern sky*/
  char wname[200];

  myfar = 8 * MAX(gemXmax,gemYmax);

  glGetFloatv(GL_DEPTH_CLEAR_VALUE,&tmpf);
  
  tmp = glutGet(GLUT_WINDOW_DEPTH_SIZE);

  glutInitWindowSize(WIDTH_3D, HEIGHT_3D);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
  sprintf(wname, "%s", gemname);
  pspWin = glutCreateWindow(wname);
  glutSetWindowTitle(wname);

  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  /*  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);*/
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);

  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  glDisable(GL_POLYGON_SMOOTH);
  
  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);

  glutDisplayFunc(refreshPsp);
  glutKeyboardFunc(keyfunc);
  glutSpecialFunc(specfunc);
}


/*  refreshPsp -- redraw perspective (3-D) view  */

static void refreshPsp(void)
{
  int w, h;
  double aspect;
  
  double m[16];

  buildSurface();

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  w = glutGet(GLUT_WINDOW_WIDTH);
  h = glutGet(GLUT_WINDOW_HEIGHT);
  aspect = (double)w / (double)h;

  gluPerspective((GLdouble)(angle/aspect),(GLdouble)aspect,
		 (GLdouble)(eyeUp ? NEAR_EYE_UP : NEAR_EYE_DOWN),
		 (GLdouble)myfar);

  glGetDoublev(GL_MODELVIEW_MATRIX,m);
  
  glGetDoublev(GL_PROJECTION_MATRIX,m);

  gluLookAt((GLdouble)eyeEmx, (GLdouble)eyeEmy, (GLdouble)eyeEmz,
	    (GLdouble)refEmx, (GLdouble)refEmy, (GLdouble)refEmz,
	    (GLdouble)0, (GLdouble)0, (GLdouble)1);

  glGetDoublev(GL_PROJECTION_MATRIX,m);

  glClearColor((GLclampf)0,(GLclampf)0,(GLclampf)1,(GLclampf)0);
  glClearDepth((GLclampd)myfar);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /*  glColor3d((GLdouble)1,(GLdouble)1,(GLdouble)1); */
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  renderPsp();


  glutSwapBuffers();

}

/* renderPsp - used by refreshPsp and outputEPS */
void renderPsp(void)
{
  emitSurface(hierarchy, FILL3D);

  if (threeDMesh) {
    glColor3f((GLdouble)0,(GLdouble)0,(GLdouble)0);
    /*    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); */
    emitSurface(hierarchy, WIRE3D);
  }
}
